/*-
 * Copyright 2022,2023 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Nils Goroll <nils.goroll@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "config.h"

#include <string.h>	// memset() for INIT_OBJ()
#include <stdlib.h>	// strtoul()

#include <cache/cache_varnishd.h>
#include <cache/cache_transport.h>
#include <cache/cache_filter.h>
#include <vcl.h>
#include <vtim.h>
#include <VSC_main.h>

#include "zlib.h"
#include "zipflow.h"

#include "vcc_zipflow_if.h"
#include "zfr_iter.h"

static void v_matchproto_(vtr_deliver_f)
zfr_deliver(struct req *req, struct boc *boc, int wantbody);
static void v_matchproto_(vtr_reembark_f)
zfr_reembark(struct worker *wrk, struct req *req);
static int v_matchproto_(vtr_minimal_response_f)
zfr_minimal_response(struct req *req, uint16_t status);

//lint -e{785}	too few initializers
static const struct transport ZIPFLOW_transport = {
	.magic =		TRANSPORT_MAGIC,
	.name =			"ZIPFLOW",
	.deliver =		zfr_deliver,
	.reembark =		zfr_reembark,
	.minimal_response =	zfr_minimal_response
};

static void __attribute__((constructor))
assert_zlib(void)
{
	assert(Z_DEFAULT_COMPRESSION == -1);	//lint !e506 const bool
}

#define WS_TOP_ALLOC_OBJ(ctx, ptr, magic) do {			\
	AN((ctx)->req);					\
	AN((ctx)->req->top);					\
	AN((ctx)->req->top->topreq);				\
	ptr = WS_Alloc((ctx)->req->top->topreq->ws,		\
	    (int)sizeof *(ptr));				\
	if ((ptr) == NULL)					\
		VRT_fail(ctx, "Out of workspace for " #magic);	\
	else							\
		INIT_OBJ(ptr, magic);				\
} while(0)

static char default_level = Z_DEFAULT_COMPRESSION;

static const void * const zipflow_request_priv = &zipflow_request_priv;
static const void * const zipflow_top_priv = &zipflow_top_priv;

struct zipflow_top;
struct zipflow_request {
	unsigned			magic;
#define ZIPFLOW_REQUEST_MAGIC		0xaa175160
	unsigned			bundle:1;
	unsigned			woken:1;	// reembark
	char				level;
	VCL_STRING			host;
	VCL_STRING			uri;
	VCL_STRING			name;
	unsigned			mode;
	VCL_TIME			atime, mtime;
	VSTAILQ_ENTRY(zipflow_request)	list;
	struct zipflow_top		*top;
	pthread_cond_t			*cond;
};

VSTAILQ_HEAD(zipflow_head, zipflow_request);

struct zipflow_top {
	unsigned		magic;
#define ZIPFLOW_TOP_MAGIC	0x5743145e
	unsigned		body_req:1;
	unsigned		body_resp:1;
	objiterate_f		*zfr_f;
	struct zfr_iter_s	*zis;
	struct zipflow_head	head;
	ZIP			*zip;
	struct req		*req;
};

static struct zipflow_top *
get_zipflow_top(VRT_CTX)
{
	struct vmod_priv *task_priv;
	struct zipflow_top *zft;

	task_priv = VRT_priv_task(ctx, zipflow_top_priv);
	if (task_priv == NULL) {
		VRT_fail(ctx, "Can not get PRIV_TASK");
		return (NULL);
	}

	if (task_priv->priv != NULL) {
		CAST_OBJ_NOTNULL(zft, task_priv->priv, ZIPFLOW_TOP_MAGIC);
		return (zft);
	}

	WS_TOP_ALLOC_OBJ(ctx, zft, ZIPFLOW_TOP_MAGIC);
	if (zft == NULL)
		return (NULL);

	VSTAILQ_INIT(&zft->head);
	return (zft);
}

static int vdp_zipflow_init(VRT_CTX, struct vdp_ctx *vdc, void **priv,
    struct objcore *oc);
static int vdp_zipflow_fini(struct vdp_ctx *vdc, void **priv);
static int vdp_zipflow_bytes(struct vdp_ctx *vdc, enum vdp_action act,
    void **priv, const void *ptr, ssize_t len);

static const struct vdp vdp_zipflow[1] = {{
	.name = "zipflow",
	.init = vdp_zipflow_init,
	.bytes = vdp_zipflow_bytes,
	.fini = vdp_zipflow_fini,
	.priv1 = NULL
}};

static int vdp_zipsub_init(VRT_CTX, struct vdp_ctx *vdc, void **priv,
    struct objcore *oc);
static int vdp_zipsub_fini(struct vdp_ctx *vdc, void **priv);
static int vdp_zipsub_bytes(struct vdp_ctx *vdc, enum vdp_action act,
    void **priv, const void *ptr, ssize_t len);

static const struct vdp vdp_zipsub[1] = {{
	.name = "zipsub",
	.init = vdp_zipsub_init,
	.bytes = vdp_zipsub_bytes,
	.fini = vdp_zipsub_fini,
	.priv1 = NULL
}};

static void
init_zipflow_request(struct zipflow_request *zfr, struct zipflow_top *zft)
{

	CHECK_OBJ_NOTNULL(zfr, ZIPFLOW_REQUEST_MAGIC);
	CHECK_OBJ_NOTNULL(zft, ZIPFLOW_TOP_MAGIC);

	VSTAILQ_INSERT_TAIL(&zft->head, zfr, list);

	zfr->bundle = 1;
	zfr->level = default_level;
	zfr->name = NULL;
	zfr->mode = 0644;
	zfr->atime = -1;
	zfr->mtime = -1;
	zfr->top = zft;
}

static struct zipflow_request *
new_zipflow_request(VRT_CTX, struct zipflow_top *zft)
{
	struct zipflow_request *zfr;

	WS_TOP_ALLOC_OBJ(ctx, zfr, ZIPFLOW_REQUEST_MAGIC);
	if (zfr == NULL)
		return (NULL);

	init_zipflow_request(zfr, zft);
	return (zfr);
}

static struct zipflow_request *
get_zipflow_request(VRT_CTX)
{
	struct vmod_priv *task_priv;
	struct zipflow_request *zfr;
	struct zipflow_top *zft;
	struct req *req;

	req = ctx->req;
	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);

	if (req->transport == &ZIPFLOW_transport) {
		CAST_OBJ_NOTNULL(zfr, req->transport_priv,
		    ZIPFLOW_REQUEST_MAGIC);
		return (zfr);
	}

	task_priv = VRT_priv_task(ctx, zipflow_request_priv);
	if (task_priv == NULL) {
		VRT_fail(ctx, "Can not get PRIV_TASK");
		return (NULL);
	}

	if (task_priv->priv != NULL) {
		CAST_OBJ_NOTNULL(zfr, task_priv->priv, ZIPFLOW_REQUEST_MAGIC);
		return (zfr);
	}

	zft = get_zipflow_top(ctx);
	if (zft == NULL)
		return (NULL);

	zfr = new_zipflow_request(ctx, zft);
	if (zfr == NULL)
		return (NULL);

	task_priv->priv = zfr;
	return (zfr);
}

/*
 * ============================================================
 * VCL interface
 */

VCL_VOID
vmod_subreq(VRT_CTX, VCL_STRING uri, VCL_STRING host)
{
	struct zipflow_request *zfr;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(ctx->method & VCL_MET_TASK_C);

	if (uri == NULL || *uri != '/') {
		VRT_fail(ctx, "subreq uri argument needs to start with /");
		return;
	}

	zfr = get_zipflow_request(ctx);
	if (zfr == NULL)
		return;

	zfr = new_zipflow_request(ctx, zfr->top);
	if (zfr == NULL)
		return;

	zfr->host = host;
	zfr->uri = uri;
}

VCL_VOID
vmod_subreqs_from_body(VRT_CTX, VCL_ENUM which, VCL_ENUM format)
{
	struct zipflow_request *zfr;
	struct zipflow_top *zft;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

	if (vmod_is_subreq(ctx)) {
		VRT_fail(ctx, "subreqs_from_body can not be "
		    "called from a subrequest");
		return;
	}

	zfr = get_zipflow_request(ctx);
	if (zfr == NULL)
		return;
	zft = zfr->top;
	CHECK_OBJ_NOTNULL(zft, ZIPFLOW_TOP_MAGIC);

	// because $Restrict
	AN(ctx->req);

	if (which == VENUM(req_body)) {
		if (ctx->req->req_body_status != BS_CACHED) {
			VRT_fail(ctx, "subreqs_from_body can only "
			    "be used with a cached request body "
			    "(see std.cache_req_body())");
			return;
		}
		zft->body_req = 1;
	}
	else if (which == VENUM(resp_body)) {
		zft->body_resp = 1;
		zfr->bundle = 0;
	} else
		WRONG("which");

	if (format == VENUM(whitespace))
		zft->zfr_f = zfr_iter_ws;
	else if (format == VENUM(form))
		zft->zfr_f = zfr_iter_form;
	else
		WRONG("format");
}

VCL_BOOL
vmod_is_subreq(VRT_CTX)
{
	struct zipflow_request *zfr;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

	zfr = get_zipflow_request(ctx);
	return (zfr != NULL && zfr->uri != NULL);
}

VCL_VOID
vmod_bundle(VRT_CTX, VCL_BOOL toggle)
{
	struct zipflow_request *zfr;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

	zfr = get_zipflow_request(ctx);
	if (zfr == NULL)
		return;

	zfr->bundle = toggle ? 1 : 0;
}

VCL_VOID
vmod_set_level(VRT_CTX, VCL_INT level)
{
	struct zipflow_request *zfr;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

	if (level < -1 || level > 9) {
		VRT_fail(ctx, "invalid level %ld", level);
		return;
	}

	if (ctx->method == VCL_MET_INIT) {
		default_level = (typeof(default_level))level;
		return;
	}
	AZ(ctx->method & VCL_MET_TASK_H);

	AN(ctx->method & VCL_MET_TASK_C);
	zfr = get_zipflow_request(ctx);
	if (zfr == NULL)
		return;
	zfr->level = (char)level;
}

VCL_VOID
vmod_meta(VRT_CTX, struct VARGS(meta) *args)
{
	struct zipflow_request *zfr = get_zipflow_request(ctx);
	unsigned long u;
	char *e;

	(void)args;
	if (zfr == NULL)
		return;

	if (args->valid_name)
		zfr->name = args->name;

	if (args->valid_mode) {
		e = NULL;
		u = strtoul(args->mode, &e, 8);
		if (u == ULONG_MAX || (e != NULL && *e != '\0')) {
			VRT_fail(ctx, ".meta() error converting mode at %s: "
			    "%d (%s)", e, errno, VAS_errtxt(errno));
		}
		else {
			assert(u < UINT_MAX);
			zfr->mode = (unsigned)u;
		}
	}
	if (args->valid_atime)
		zfr->atime = args->atime;
	if (args->valid_mtime)
		zfr->mtime = args->mtime;
}

static const char *
get_url(VRT_CTX)
{
	const char *url;
	VCL_HTTP http = NULL;

	if (ctx->http_bereq != NULL)
		http = ctx->http_bereq;
	else if (ctx->http_req != NULL)
		http = ctx->http_req;
	else
		WRONG("No http headers");

	AN(http);
	url = http->hd[HTTP_HDR_URL].b;
	AN(url);
	AN(*url);
	return (url);
}

static const char *
default_name(VRT_CTX)
{
	const char *p, *url = get_url(ctx);
	ssize_t l;
	char *q;

	url = strrchr(url, '/');
	AN(url);
	url++;

	if (*url == '\0' || *url == '?')
		return ("unnamed_file");

	p = strchr(url, '?');
	if (p == NULL)
		return (url);

	assert(p > url);
	l = p - url;
	l++;
	q = WS_Copy(ctx->ws, url, (int)l);
	if (q == NULL)
		return ("unnamed_file");
	q[l] = '\0';
	return (q);
}

static VCL_TIME
get_lm(VRT_CTX)
{
	const char *p;
	VCL_TIME lm;
	VCL_HTTP http;

	if (ctx->http_beresp != NULL)
		http = ctx->http_beresp;
	else if (ctx->http_resp != NULL)
		http = ctx->http_resp;
	else
		return (ctx->now);

	if (!http_GetHdr(http, H_Last_Modified, &p))
		return (ctx->now);

	lm = VTIM_parse(p);
	return (lm ? lm : ctx->now);
}

static void
fill_meta(VRT_CTX, struct zipflow_request *zfr)
{

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(zfr, ZIPFLOW_REQUEST_MAGIC);

	if (zfr->name == NULL)
		zfr->name = default_name(ctx);

	if (zfr->atime < 0)
		zfr->atime = ctx->now;
	if (zfr->mtime < 0)
		zfr->mtime = get_lm(ctx);
	VSLb(ctx->vsl, SLT_Debug, "%s %o %f %f", zfr->name, zfr->mode,
	    zfr->atime, zfr->mtime);
}

int
vmod_event(VRT_CTX, struct vmod_priv *priv, enum vcl_event_e e)
{
	(void) priv;

	switch(e) {
	case VCL_EVENT_LOAD:
		if (VRT_AddFilter(ctx, NULL, vdp_zipflow) != NULL)
			return (1);
		break;
	case VCL_EVENT_DISCARD:
		VRT_RemoveFilter(ctx, NULL, vdp_zipflow);
		break;
	case VCL_EVENT_WARM:
	case VCL_EVENT_COLD:
		break;
	default:
		WRONG("illegal event enum");
	}
	return (0);
}

/*
 * ============================================================
 * Iterators: Iterate over a body to create sub requests
 */

static void
zfr_include(struct req *preq, struct zipflow_request *zfr);

static void
zfr_iter_include(void *priv, const char *u, const char *h)
{
	struct zipflow_top *zft;
	struct zipflow_request *zfr, this[1];

	CAST_OBJ_NOTNULL(zft, priv, ZIPFLOW_TOP_MAGIC);

	while ((zfr = VSTAILQ_FIRST(&zft->head)) != NULL)
		zfr_include(zft->req, zfr);

	INIT_OBJ(this, ZIPFLOW_REQUEST_MAGIC);
	init_zipflow_request(this, zft);
	this->host = h;
	this->uri = u;
	assert(this == VSTAILQ_FIRST(&zft->head));

	while ((zfr = VSTAILQ_FIRST(&zft->head)) != NULL)
		zfr_include(zft->req, zfr);
}

static void
zft_iter_req_body(struct zipflow_top *zft)
{
	struct zfr_iter_s zis[1];
	struct req *req;

	CHECK_OBJ_NOTNULL(zft, ZIPFLOW_TOP_MAGIC);
	req = zft->req;
	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);

	INIT_OBJ(zis, ZFR_ITER_MAGIC);
	zis->zft = zft;
	zis->func = zfr_iter_include;
	zis->priv = zft;

	if (zft->body_req) {
		AN(zft->zfr_f);
		(void) VRB_Iterate(req->wrk, req->vsl, req,
		    zft->zfr_f, (void *)zis);
		zfr_iter_fini(zis);
	}
}

/*
 * ============================================================
 * VDP
 */


static int
vdp_zipflow_put(void *priv, void const *ptr, size_t len)
{

	return (VDP_bytes(priv, ptr == NULL ? VDP_END : VDP_FLUSH,
	    ptr, (ssize_t)len));
}

static void
vdp_zipflow_log(void *vsl, char *msg)
{

	AN(vsl);
	AN(msg);
	VSLb(vsl, SLT_Error, "zipflow: %s", msg);
	free(msg);
}

static struct zipflow_top *
zipsub_take(const struct zipflow_request *zfr)
{
	struct zipflow_top *zft;

	zft = zfr->top;
	CHECK_OBJ_NOTNULL(zft, ZIPFLOW_TOP_MAGIC);

	assert(zfr == VSTAILQ_FIRST(&zft->head));
	VSTAILQ_REMOVE_HEAD(&zft->head, list);

	AN(zft->zip);
	return (zft);
}

static int
vdp_zipsub_init(VRT_CTX, struct vdp_ctx *vdc, void **priv, struct objcore *oc)
{
	struct zipflow_request *zfr;
	struct zipflow_top *zft;
	struct req *req;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	(void) vdc;
	AN(vdc);
	AN(priv);
	(void) priv;
	CAST_OBJ_NOTNULL(zfr, *priv, ZIPFLOW_REQUEST_MAGIC);
	(void) oc;

	zft = zipsub_take(zfr);

	req = vdc->req;
	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);

	if (req->resp_len == 0)
		zfr->bundle = 0;

	if (zfr->bundle) {
		fill_meta(ctx, zfr);
		AZ(zip_meta(zft->zip, zfr->name, 3, zfr->mode,
			(uint32_t)zfr->atime, (uint32_t)zfr->mtime));
	}

	RFC2616_Weaken_Etag(req->resp);
	if (req->resp_len != 0)
		req->resp_len = -1;

	return (0);
}

static int
vdp_zipflow_init(VRT_CTX, struct vdp_ctx *vdc, void **priv, struct objcore *oc)
{
	struct zipflow_request *zfr = get_zipflow_request(ctx);
	struct zipflow_top *zft;

	if (zfr == NULL)
		return (1);

	zft = zfr->top;
	CHECK_OBJ_NOTNULL(zft, ZIPFLOW_TOP_MAGIC);
	CHECK_OBJ_NOTNULL(vdc, VDP_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(vdc->req, REQ_MAGIC);

	if (zft->req) {
		VSLb(vdc->vsl, SLT_Error, "zipflow: can't be nested");
		return (-1);
	}

	if (vdc->req->resp_len == 0) {
		VSLb(vdc->vsl, SLT_Error, "zipflow: need body bytes");
		return (-1);
	}

	if (zft->body_resp) {
		zfr->bundle = 0;
		WS_TOP_ALLOC_OBJ(ctx, zft->zis, ZFR_ITER_MAGIC);
		if (zft->zis == NULL) {
			VSLb(vdc->vsl, SLT_Error, "zipflow: workspace "
			    "overflow for body_resp");
			return (-1);
		}
		zft->zis->zft = zft;
		zft->zis->func = zfr_iter_include;
		zft->zis->priv = zft;
	}

	AZ(zft->req);
	zft->req = vdc->req;
	AZ(zft->zip);
	zft->zip = zip_pipe(vdc, vdp_zipflow_put, zfr->level);
	if (zft->zip == NULL) {
		VSLb(vdc->vsl, SLT_Error, "zipflow: zip_pipe failed");
		return (-1);
	}
	AZ(zip_log(zft->zip, vdc->vsl, vdp_zipflow_log));

	AN(priv);
	AZ(*priv);
	*priv = zfr;

	return (vdp_zipsub_init(ctx, vdc, priv, oc));
}

static void
zipsub_fini(struct zipflow_request *zfr)
{
	CHECK_OBJ_NOTNULL(zfr, ZIPFLOW_REQUEST_MAGIC);
	memset(zfr, 0, sizeof *zfr);
}

static int
vdp_zipsub_fini(struct vdp_ctx *vdc, void **priv)
{
	struct zipflow_request *zfr;

	(void) vdc;

	AN(priv);
	zfr = *priv;
	*priv = NULL;
	if (zfr != NULL)
		zipsub_fini(zfr);
	return (0);
}

static int
vdp_zipflow_fini(struct vdp_ctx *vdc, void **priv)
{
	struct zipflow_request *zfr;
	struct zipflow_top *zft;
	int r;

	(void) vdc;
	AN(priv);
	zfr = *priv;
	*priv = NULL;
	if (zfr == NULL)
		return (0);

	CHECK_OBJ(zfr, ZIPFLOW_REQUEST_MAGIC);
	zft = zfr->top;
	CHECK_OBJ_NOTNULL(zft, ZIPFLOW_TOP_MAGIC);
	r = zip_close(zft->zip);
	if (r)
		VSLb(vdc->vsl, SLT_Error, "zip_close returned %d", r);
	memset(zfr, 0, sizeof *zfr);

	return (0);
}

static int
vdp_zipsub_bytes(struct vdp_ctx *vdc, enum vdp_action act, void **priv,
    const void *ptr, ssize_t len)
{
	struct zipflow_request *zfr;
	struct zipflow_top *zft;
	int r;

	(void) vdc;
	(void) priv;
	AN(priv);
	CAST_OBJ_NOTNULL(zfr, *priv, ZIPFLOW_REQUEST_MAGIC);
	zft = zfr->top;
	CHECK_OBJ_NOTNULL(zft, ZIPFLOW_TOP_MAGIC);
	if (zfr->bundle == 0)
		return (0);

	assert(len >= 0);
	r = zip_data(zft->zip, ptr, (size_t)len, act == VDP_END ? 1 : 0);
	if (r) {
		VSLb(vdc->vsl, SLT_Error, "zip_data returned %d", r);
		return (-1);
	}

	return (0);
}

static int
vdp_zipflow_bytes(struct vdp_ctx *vdc, enum vdp_action act, void **priv,
    const void *ptr, ssize_t len)
{
	struct zipflow_request *zfr;
	struct zipflow_top *zft;
	int r, c;

	AN(priv);
	CAST_OBJ_NOTNULL(zfr, *priv, ZIPFLOW_REQUEST_MAGIC);
	zft = zfr->top;

	if (zft->zis != NULL) {
		AN(zft->zfr_f);
		r = zft->zfr_f(zft->zis, (act == VDP_END) ? OBJ_ITER_END : 0,
		    ptr, len);
	}
	else
		r = vdp_zipsub_bytes(vdc, act, priv, ptr, len);

	if (r == 0 && act != VDP_END)
		return (r);

	*priv = NULL;
	if (zft->zis != NULL) {
		zfr_iter_fini(zft->zis);
		zft->zis = NULL;
	}

	memset(zfr, 0, sizeof *zfr);
	CHECK_OBJ_NOTNULL(zft, ZIPFLOW_TOP_MAGIC);

	if (r == 0) {
		while ((zfr = VSTAILQ_FIRST(&zft->head)) != NULL)
			zfr_include(zft->req, zfr);
	}

	zft_iter_req_body(zft);

	c = zip_close(zft->zip);
	if (c)
		VSLb(vdc->vsl, SLT_Error, "zip_close returned %d", r);
	if (r == 0)
		r = c;
	return (r);
}


static void v_matchproto_(vtr_reembark_f)
zfr_reembark(struct worker *wrk, struct req *req)
{
	struct zipflow_request *zfr;

	(void)wrk;
	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
	CAST_OBJ_NOTNULL(zfr, req->transport_priv, ZIPFLOW_REQUEST_MAGIC);
	AN(zfr->cond);
	AN(req->sp);
	Lck_Lock(&req->sp->mtx);
	zfr->woken = 1;
	PTOK(pthread_cond_signal(zfr->cond));
	Lck_Unlock(&req->sp->mtx);
}

static int v_matchproto_(vtr_minimal_response_f)
zfr_minimal_response(struct req *req, uint16_t status)
{
	(void)req;
	(void)status;
	WRONG("zfr_minimal_response should not be called");
	return (0);
}

/*--------------------------------------------------------------------
 * basically varnish-cache ESI
*/

static void
zfr_include(struct req *preq, struct zipflow_request *zfr)
{
	struct worker *wrk;
	struct sess *sp;
	struct req *req;
	enum req_fsm_nxt s;
	const char *h;

	CHECK_OBJ_NOTNULL(preq, REQ_MAGIC);
	CHECK_OBJ_NOTNULL(preq->top, REQTOP_MAGIC);
	sp = preq->sp;
	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
	CHECK_OBJ_NOTNULL(zfr, ZIPFLOW_REQUEST_MAGIC);
	wrk = preq->wrk;

	if (preq->esi_level >= cache_param->max_esi_depth) {
		VSLb(preq->vsl, SLT_VCL_Error,
		    "ESI depth limit reached (param max_esi_depth = %u)",
		    cache_param->max_esi_depth);
		return;
	}

	req = Req_New(sp);
	AN(req);
	THR_SetRequest(req);
	assert(IS_NO_VXID(req->vsl->wid));
	req->vsl->wid = VXID_Get(wrk, VSL_CLIENTMARKER);

	wrk->stats->esi_req++;
	req->esi_level = preq->esi_level + 1;

	VSLb(req->vsl, SLT_Begin, "req %ju esi %u",
	    (uintmax_t)VXID(preq->vsl->wid), req->esi_level);
	VSLb(preq->vsl, SLT_Link, "req %ju esi %u",
	    (uintmax_t)VXID(req->vsl->wid), req->esi_level);

	VSLb_ts_req(req, "Start", W_TIM_real(wrk));

	memset(req->top, 0, sizeof *req->top);
	req->top = preq->top;

	HTTP_Setup(req->http, req->ws, req->vsl, SLT_ReqMethod);
	HTTP_Dup(req->http, preq->http0);

	http_SetH(req->http, HTTP_HDR_URL, zfr->uri);
	if (zfr->host != NULL && *zfr->host != '\0')  {
		http_Unset(req->http, H_Host);
		h = WS_Printf(req->ws, "Host: %s", zfr->host);
		http_SetHeader(req->http, h);
	}

	http_ForceField(req->http, HTTP_HDR_METHOD, "GET");
	http_ForceField(req->http, HTTP_HDR_PROTO, "HTTP/1.1");

	/* Don't allow conditionals, we can't use a 304 */
	http_Unset(req->http, H_If_Modified_Since);
	http_Unset(req->http, H_If_None_Match);

	/* Don't allow Range */
	http_Unset(req->http, H_Range);

	// XXX handle gzip when zipflow supports it
	http_Unset(req->http, H_Accept_Encoding);

	/* Client content already taken care of */
	http_Unset(req->http, H_Content_Length);
	http_Unset(req->http, H_Transfer_Encoding);
	req->req_body_status = BS_NONE;

	AZ(req->vcl);
	AN(req->top);
	if (req->top->vcl0)
		req->vcl = req->top->vcl0;
	else
		req->vcl = preq->vcl;
	VCL_Ref(req->vcl);

	assert(req->req_step == R_STP_TRANSPORT);
	req->t_req = preq->t_req;

	req->transport = &ZIPFLOW_transport;
	req->transport_priv = zfr;

	VCL_TaskEnter(req->privs);

	AZ(zfr->cond);
	zfr->cond = &wrk->cond;

	//lint -e{716} while(1)
	while (1) {
		CNT_Embark(wrk, req);
		zfr->woken = 0;
		s = CNT_Request(req);
		if (s == REQ_FSM_DONE)
			break;
		DSL(DBG_WAITINGLIST, req->vsl->wid,
		    "waiting for zipflow (%d)", (int)s);
		assert(s == REQ_FSM_DISEMBARK);

		AN(zfr->cond);
		Lck_Lock(&sp->mtx);
		if (!zfr->woken)
			(void)Lck_CondWait(zfr->cond, &sp->mtx);
		Lck_Unlock(&sp->mtx);
		AZ(req->wrk);
	}

	VCL_Rel(&req->vcl);

	req->wrk = NULL;
	THR_SetRequest(preq);

	Req_Cleanup(sp, wrk, req);
	Req_Release(req);
}

static void v_matchproto_(vtr_deliver_f)
zfr_deliver(struct req *req, struct boc *boc, int wantbody)
{
	int i;
	const char *p;
	uint16_t status;
	struct vrt_ctx ctx[1];

	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
	CHECK_OBJ_ORNULL(boc, BOC_MAGIC);
	CHECK_OBJ_NOTNULL(req->objcore, OBJCORE_MAGIC);

	status = req->resp->status % 1000;

	if (wantbody == 0 ||
	    status != 200 ||
	    (boc == NULL && ObjGetLen(req->wrk, req->objcore) == 0)) {
		(void) zipsub_take(req->transport_priv);
		zipsub_fini(req->transport_priv);
		return;
	}

	AZ(http_GetHdr(req->resp, H_Content_Encoding, &p));

	INIT_OBJ(ctx, VRT_CTX_MAGIC);
	VCL_Req2Ctx(ctx, req);

	i = VDP_Push(ctx, req->vdc, req->ws, vdp_zipsub, req->transport_priv);

	if (i == 0) {
		i = VDP_DeliverObj(req->vdc, req->objcore);
	} else {
		VSLb(req->vsl, SLT_Error, "Failure to push ZIP processors");
		req->doclose = SC_OVERLOAD;
	}

	if (i && req->doclose == SC_NULL)
		req->doclose = SC_REM_CLOSE;

	req->acct.resp_bodybytes += VDP_Close(req->vdc, req->objcore, boc);

#if 0
	if (i && !ecx->incl_cont) {
		req->top->topreq->vdc->retval = -1;
		req->top->topreq->doclose = req->doclose;
	}
#endif
}
