/*-
 * Copyright 2022,2023 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Nils Goroll <nils.goroll@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * iterate over a body and call callback for anything which looks like a url
 */

#include "config.h"

#include <string.h>	// memset() for INIT_OBJ()
#include <stdlib.h>	// strtoul()
#include <stdio.h>

#include <cache/cache.h>
#include <vct.h>

#include "zfr_iter.h"

/* This function exists because flexelint does not understand (p += strlen(s))
 * in place
 */

static inline int
pincr(const char **p, size_t l)
{
	*p += l;
	return (1);
}

#define tok(p, e, s) (							\
	    pdiff(p,e) >= strlen(s) &&					\
	    strncmp(p, s, strlen(s)) == 0 &&				\
	    pincr(&p, strlen(s))					\
	    )

#ifdef TEST_DRIVER
static const char *DBG_begin;
#define DBG_BEGIN(p) DBG_begin = (p)

#define DBG_INVALID(p, why) do {					\
		printf("invalid " why " >%.*s<\n",			\
		    (int)pdiff(DBG_begin, p), DBG_begin);		\
	} while (0)

#define DBG_VALID(p) \
	printf("valid >%.*s<\n", (int)pdiff(DBG_begin, p), DBG_begin)
#else
#define DBG_BEGIN(p)		(void)0
#define DBG_INVALID(p, why)	(void)0
#define DBG_VALID(p)		(void)0
#endif

/* iterate over whitespace separated input */
int
zfr_iter_ws(void *priv, unsigned flush, const void *ptr, ssize_t alen)
{
	const char *pp, *p, *e, *h, *u;
	struct zfr_iter_s *zis;
	char *hh, *uu;
	size_t l;

	CAST_OBJ_NOTNULL(zis, priv, ZFR_ITER_MAGIC);

	AZ(zis->end_seen);
	zis->end_seen = ((flush & OBJ_ITER_END) != 0);

#ifdef DEBUG
	printf(
	    "buf %zu >%.*s<\t"
	    "ptr %zu >%.*s<\n",
	    zis->len, (int)zis->len, zis->buf,
	    len, (int)len, (const char *)ptr);
#endif

	if (zis->len > 0)
		AN(zis->buf);

	if (alen == 0 && zis->len == 0)
		return (0);

	assert(alen >= 0);
	l = (size_t) alen;

	if (zis->len == 0)
		p = ptr;
	else {
		zis->buf = realloc(zis->buf, zis->len + l);
		AN(zis->buf);
		memcpy(zis->buf + zis->len, ptr, l);
		zis->len += l;

		p = zis->buf;
		l = zis->len;
	}
	AN(p);

	pp = p;
	e = p + l;
	while (p < e) {
		if (vct_islws(*p)) {
			p++;
			continue;
		}
		DBG_BEGIN(p);
		h = NULL;
		if (tok(p, e, "//") ||
		    tok(p, e, "http://") ||
		    tok(p, e, "https://")) {
			// sufficient to keep "//" in buffer
			pp = p - 2;
			h = p;
			while (p < e && !vct_islws(*p) &&  *p != '/')
				p++;
		}
		u = p;

		// skip over whatever non-whitespace
		while (p < e && !vct_islws(*p))
			p++;

		// continue in next chunk
		if (p == e && (! (flush & OBJ_ITER_END)))
			break;

		// token complete, checkpoint
		pp = p;

		if (*u != '/') {
			DBG_INVALID(p, "non-url");
			continue;
		}

		if (h != NULL && h == u) {
			DBG_INVALID(p, "no-host");
			continue;
		}

		/* match! */
		DBG_VALID(p);
		hh = NULL;
		if (h) {
			assert(u > h);
			hh = strndup(h, pdiff(h, u));
			AN(hh);
		}
		assert(p > u);
		uu = strndup(u, pdiff(u, p));
		AN(uu);
		AN(zis->func);
		zis->func(zis->priv, uu, hh);
		free(hh);
		free(uu);
	}

	assert(e >= pp);
	l = pdiff(pp, e);
	if (flush & OBJ_ITER_END || l == 0) {
		zis->len = 0;
		return (0);
	}
	if (zis->len)
		memmove(zis->buf, pp, l);
	else {
		zis->buf = realloc(zis->buf, l);
		memcpy(zis->buf, pp, l);
	}
	zis->len = l;
	return (0);
}

void
zfr_iter_fini(struct zfr_iter_s *zis)
{

	CHECK_OBJ_NOTNULL(zis, ZFR_ITER_MAGIC);

	if (zis->len > 0 && zis->end_seen == 0)
		(void) zfr_iter_ws(zis, OBJ_ITER_END, "", (ssize_t)0);
	free(zis->buf);
	zis->buf = NULL;
	zis->len = 0;
}

/* iterate over application/x-www-form-urlencoded
 * by converting to whitespace format and passing to zfr_iter_ws()
 */

int
zfr_iter_form(void *priv, unsigned flush, const void *ptr, ssize_t alen)
{
	struct zfr_iter_form_s *zfrm;
	struct zfr_iter_s *zis;
	const char *pp, *p, *e;
	unsigned end;
	size_t l;
	int r = 0;

	CAST_OBJ_NOTNULL(zis, priv, ZFR_ITER_MAGIC);
	zfrm = &zis->form;

	assert(alen >= 0);
	l = (size_t) alen;
	pp = ptr;
	e = pp + l;

	while (pp < e) {
		switch (zfrm->state) {
		case FRMST_KEY:
			if (*pp != '=')
				break;
			zfrm->state = FRMST_VAL;
			break;
		case FRMST_VAL:
			p = pp;
			while (p < e &&
			    *p != '%' &&
			    *p != '&')
				p++;
			end = 0;
			if (p != pp) {
				end = (p == e) ? flush : 0;
				r = zfr_iter_ws(priv, end,
				    pp, (ssize_t)pdiff(pp, p));
				if (r)
					return (r);
				pp = p;
			}
			if (pp == e)
				continue;
			else if (*pp == '%')
				zfrm->state = FRMST_PERC;
			else if (end)
				break;
			else if (*pp == '&') {
				r = zfr_iter_ws(priv, 0,
				    " ", (ssize_t)1);
				if (r)
					return (r);
				zfrm->state = FRMST_KEY;
			}
			else
				WRONG("impossible form val");
			break;
		case FRMST_PERC:
			if (*pp >= '0' && *pp <= '9')
				zfrm->hex = (unsigned char)*pp - '0';
			else if (*pp >= 'a' && *pp <= 'f')
				zfrm->hex = 0xa + (unsigned char)*pp - 'a';
			else if (*pp >= 'A' && *pp <= 'F')
				zfrm->hex = 0xa + (unsigned char)*pp - 'A';
			else
				return (-2);
			zfrm->hex <<= 4;
			zfrm->state = FRMST_PERC1;
			break;
		case FRMST_PERC1:
			if (*pp >= '0' && *pp <= '9')
				zfrm->hex |= *pp - '0';
			else if (*pp >= 'a' && *pp <= 'f')
				zfrm->hex |= 0xa + *pp - 'a';
			else if (*pp >= 'A' && *pp <= 'F')
				zfrm->hex |= 0xa + *pp - 'A';
			else
				return (-3);
			end = (pp == e) ? flush : 0;
			r = zfr_iter_ws(priv, end,
			    &zfrm->hex, (ssize_t)1);
			if (r)
				return (r);
			zfrm->state = FRMST_VAL;
			break;
		default:
			WRONG("zfrm->state");
		}
		pp++;
	}
	assert(pp == e);
	AZ(r);
	if ((flush & OBJ_ITER_END) != 0 && zis->end_seen == 0)
		r = zfr_iter_ws(priv, flush, "", (ssize_t)0);
	return (r);
}

#ifdef TEST_DRIVER
#include <stdio.h>

struct expect {
	const char *eu, *u, *h, **pfx;
};

/* testcase whitespace
 * -------------
 */

static const char *pfx_ws_host[7] = {
	"https://",
	"http://",
	"//",
	"  https://",
	" http://",
	"   //",
	NULL
};

static const char *pfx_ws_nohost[2] = {
	"",
	NULL
};

// incomplete tokens to skip
static const char *invalid_ws[10] = {
	// lone prefixes
	"https://",
	"http:///",
	"//",
	"///",
	"///a",
	"//a",
	// partials
	"http",
	"invalid",
	"a",
	NULL
};

static struct expect tc_ws[9] = {
	{ .u = "/url", .h = "host", .pfx = pfx_ws_host },
	{ .u = "/",    .h = "host", .pfx = pfx_ws_host },
	{ .u = "///",  .h = "ho-t", .pfx = pfx_ws_host },
	{ .u = "/",    .h = "h",    .pfx = pfx_ws_host },
	{ .u = "///",  .h = "h",    .pfx = pfx_ws_host },
	{ .u = "/url", .h = NULL,   .pfx = pfx_ws_nohost },
	{ .u = "/",    .h = NULL,   .pfx = pfx_ws_nohost },
	{ .u = "/a",   .h = NULL,   .pfx = pfx_ws_nohost },
	{ NULL }
};

/* testcase form
 * -------------
 */
static const char *pfx_form_host[7] = {
	"https%3a%2F%2F",
	"http%3a%2F%2F",
	"%2F%2F",
	"  https%3a%2F%2F",
	" http%3a%2F%2F",
	"   %2F%2F",
	NULL
};

static const char *pfx_form_nohost[2] = {
	"",
	NULL
};

// incomplete tokens to skip
static const char *invalid_form[10] = {
	// lone prefixes
	"https%3a%2F%2F",
	"http%3a%2F%2F%2F",
	"%2F%2F",
	"%2F%2F%2F",
	"%2F%2F%2Fa",
	"%2F%2Fa",
	// partials
	"http",
	"invalid",
	"a",
	NULL
};

static struct expect tc_form[10] = {
	{ .eu = "/url", .u = "/url",      .h = "host", .pfx = pfx_form_host },
	{ .eu = "/url", .u = "%2Furl",    .h = "host", .pfx = pfx_form_host },
	{ .eu = "/",    .u = "%2F",       .h = "host", .pfx = pfx_form_host },
	{ .eu = "///",  .u = "%2F%2F%2F", .h = "ho-t", .pfx = pfx_form_host },
	{ .eu = "/",    .u = "%2F",       .h = "h",    .pfx = pfx_form_host },
	{ .eu = "///",  .u = "%2F%2F%2F", .h = "h",    .pfx = pfx_form_host },
	{ .eu = "/url", .u = "%2Furl",    .h = NULL,   .pfx = pfx_form_nohost },
	{ .eu = "/",    .u = "%2F",       .h = NULL,   .pfx = pfx_form_nohost },
	{ .eu = "/a",   .u = "%2Fa",      .h = NULL,   .pfx = pfx_form_nohost },
	{ NULL }
};

struct cb_want_priv {
	unsigned		magic;
#define CBW_MAGIC		0x79d8b905
	unsigned		count;
	const struct expect	*e;
};

static void
cb_not(void *priv, const char *u, const char *h)
{
	WRONG("not to be called");
}

typedef void want_f(void *priv, const char *u, const char *h);

static void
cb_want_ws(void *priv, const char *u, const char *h)
{
	struct cb_want_priv *p;
	const struct expect *e;

	CAST_OBJ_NOTNULL(p, priv, CBW_MAGIC);
	e = p->e;

	if (e->u == NULL)
		AZ(u);
	else {
		AN(u);
		AZ(strcmp(e->u, u));
	}

	if (e->h == NULL)
		AZ(h);
	else {
		AN(h);
		AZ(strcmp(e->h, h));
	}

	p->count++;
}

static void
cb_want_form(void *priv, const char *u, const char *h)
{
	struct cb_want_priv *p;
	const struct expect *e;

	CAST_OBJ_NOTNULL(p, priv, CBW_MAGIC);
	e = p->e;

	if (e->eu == NULL)
		AZ(u);
	else {
		AN(u);
		AZ(strcmp(e->eu, u));
	}

	if (e->h == NULL)
		AZ(h);
	else {
		AN(h);
		AZ(strcmp(e->h, h));
	}

	p->count++;
}

static void
t_steps(objiterate_f *f, want_f *wf,
    const struct expect *want, unsigned n, const char *s)
{
	struct cb_want_priv wp[1];
	struct zfr_iter_s zis[1];
	size_t l, step;
	const char *t;

	INIT_OBJ(zis, ZFR_ITER_MAGIC);
	INIT_OBJ(wp, CBW_MAGIC);

	wp->e = want;
	zis->priv = wp;
	zis->func = wf;
	for (step = 1; step < strlen(s); step++) {
		wp->count = 0;
		zis->end_seen = 0;
		for (t = s, l = strlen(t);
		     l > step;
		     l -= step, t += step)
			f(zis, 0, t, step);
		AN(l);
		f(zis, OBJ_ITER_END, t, l);
		AN(zis->end_seen);
		assert(wp->count == n);
	}
}

int
main(void) {
	const struct expect *want;
	struct zfr_iter_s zis[1];
	const char **p, *t, **inv = invalid_ws;
	char s[256], u[64];
	unsigned n;
	size_t l;

	// base test of tok()
	strcpy(s, "https://");
	t = s;
	assert(tok(t, strchr(s, '\0'), "https://"));

	for (want = tc_ws; want->u != NULL; want++) {
		INIT_OBJ(zis, ZFR_ITER_MAGIC);
		zis->priv = NULL;
		zis->func = cb_not;
		zfr_iter_ws(zis, 0, "", 0);

		n = 0;
		*s = '\0';
		for (p = want->pfx; *p != NULL; p++, inv++) {
			AN(inv);
			if (*inv == NULL)
				inv = invalid_ws;
			AN(p);
			AN(*p);
			if (**p == '\0') {
				l = snprintf(u, sizeof u, " %s\t%s",
				    *inv, want->u);
			}
			else {
				l = snprintf(u, sizeof u, " %s\t%s%s%s",
				    *inv, *p, want->h, want->u);
			}
			assert(l < sizeof u);
			inv++;

			if (sizeof s - strlen(s) - 1 < l)
				break;
			(void) strcat(s, u);
			n++;

			t_steps(zfr_iter_ws, cb_want_ws, want, n, s);

			if (*inv == NULL)
				continue;

			if (sizeof s - strlen(s) - 2 < strlen(*inv))
				break;
			(void) strcat(s, "\r");
			(void) strcat(s, *inv);
			t_steps(zfr_iter_ws, cb_want_ws, want, n, s);
			inv++;
		}
	}
	zfr_iter_fini(zis);

	// form
	inv = invalid_form;
	for (want = tc_form; want->u != NULL; want++) {
		INIT_OBJ(zis, ZFR_ITER_MAGIC);
		zis->priv = NULL;
		zis->func = cb_not;
		zfr_iter_ws(zis, 0, "", 0);

		n = 0;
		*s = '\0';
		for (p = want->pfx; *p != NULL; p++, inv++) {
			AN(inv);
			if (*inv == NULL)
				inv = invalid_form;
			AN(p);
			AN(*p);
			if (**p == '\0') {
				l = snprintf(u, sizeof u, " %s\tfoo=%s&",
				    *inv, want->u);
			}
			else {
				l = snprintf(u, sizeof u, " %s&bar=%s%s%s&",
				    *inv, *p, want->h, want->u);
			}
			assert(l < sizeof u);
			inv++;

			if (sizeof s - strlen(s) - 1 < l)
				break;
			(void) strcat(s, u);
			n++;

			t_steps(zfr_iter_form, cb_want_form, want, n, s);

			if (*inv == NULL)
				continue;

			if (sizeof s - strlen(s) - 2 < strlen(*inv))
				break;
			(void) strcat(s, "\r");
			(void) strcat(s, *inv);
			t_steps(zfr_iter_form, cb_want_form, want, n, s);
			inv++;
		}
	}

	zfr_iter_fini(zis);
	printf("OK\n");

	return (0);
}
#endif
